/*
  PuttyAgent plugin for KeePass 1
  Copyright (C) 2011 Nikolaus Hammler <nikolaus@hammler.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*
  The code in this file implements the agent functionality and is derived
  from the original sources of Pageant.
*/

#include "StdAfx.h"
#include "KeyHandler.h"

CKeyHandler::CKeyHandler()
{
	rsakeys = newtree234(cmpkeys_rsa);
	ssh2keys = newtree234(cmpkeys_ssh2);
}

CKeyHandler::~CKeyHandler()
{
	DeleteKeys();
	freetree234(rsakeys);
	freetree234(ssh2keys);
}

int CKeyHandler::cmpkeys_rsa(void *av, void *bv)
{
    struct RSAKey *a = (struct RSAKey *) av;
    struct RSAKey *b = (struct RSAKey *) bv;
    Bignum am, bm;
    int alen, blen;

    am = a->modulus;
    bm = b->modulus;
    /*
     * Compare by length of moduli.
     */
    alen = bignum_bitcount(am);
    blen = bignum_bitcount(bm);
    if (alen > blen)
	return +1;
    else if (alen < blen)
	return -1;
    /*
     * Now compare by moduli themselves.
     */
    alen = (alen + 7) / 8;	       /* byte count */
    while (alen-- > 0) {
	int abyte, bbyte;
	abyte = bignum_byte(am, alen);
	bbyte = bignum_byte(bm, alen);
	if (abyte > bbyte)
	    return +1;
	else if (abyte < bbyte)
	    return -1;
    }
    /*
     * Give up.
     */
    return 0;
}

int CKeyHandler::cmpkeys_ssh2(void *av, void *bv)
{
    struct ssh2_userkey *a = (struct ssh2_userkey *) av;
    struct ssh2_userkey *b = (struct ssh2_userkey *) bv;
    int i;
    int alen, blen;
    unsigned char *ablob, *bblob;
    int c;

    /*
     * Compare purely by public blob.
     */
    ablob = a->alg->public_blob(a->data, &alen);
    bblob = b->alg->public_blob(b->data, &blen);

    c = 0;
    for (i = 0; i < alen && i < blen; i++) {
	if (ablob[i] < bblob[i]) {
	    c = -1;
	    break;
	} else if (ablob[i] > bblob[i]) {
	    c = +1;
	    break;
	}
    }
    if (c == 0 && i < alen)
	c = +1;			       /* a is longer */
    if (c == 0 && i < blen)
	c = -1;			       /* a is longer */

    sfree(ablob);
    sfree(bblob);

    return c;
}

int CKeyHandler::cmpkeys_ssh2_asymm(void *av, void *bv)
{
    struct BLOBSTRUCT *a = (struct BLOBSTRUCT *) av;
    struct ssh2_userkey *b = (struct ssh2_userkey *) bv;
    int i;
    int alen, blen;
    unsigned char *ablob, *bblob;
    int c;

    /*
     * Compare purely by public blob.
     */
    ablob = a->blob;
    alen = a->len;
    bblob = b->alg->public_blob(b->data, &blen);

    c = 0;
    for (i = 0; i < alen && i < blen; i++) {
	if (ablob[i] < bblob[i]) {
	    c = -1;
	    break;
	} else if (ablob[i] > bblob[i]) {
	    c = +1;
	    break;
	}
    }
    if (c == 0 && i < alen)
	c = +1;			       /* a is longer */
    if (c == 0 && i < blen)
	c = -1;			       /* a is longer */

    sfree(bblob);

    return c;
}

void *CKeyHandler::make_keylist1(int *length)
{
    int i, nkeys, len;
    struct RSAKey *key;
    unsigned char *blob, *p, *ret;
    int bloblen;

    /*
     * Count up the number and length of keys we hold.
     */
    len = 4;
    nkeys = 0;
    for (i = 0; NULL != (key = (RSAKey*)index234(rsakeys, i)); i++) {
	nkeys++;
	blob = rsa_public_blob(key, &bloblen);
	len += bloblen;
	sfree(blob);
	len += 4 + strlen(key->comment);
    }

    /* Allocate the buffer. */
    p = ret = snewn(len, unsigned char);
    if (length) *length = len;

    PUT_32BIT(p, nkeys);
    p += 4;
    for (i = 0; NULL != (key = (RSAKey*)index234(rsakeys, i)); i++) {
	blob = rsa_public_blob(key, &bloblen);
	memcpy(p, blob, bloblen);
	p += bloblen;
	sfree(blob);
	PUT_32BIT(p, strlen(key->comment));
	memcpy(p + 4, key->comment, strlen(key->comment));
	p += 4 + strlen(key->comment);
    }

    //assert(p - ret == len);
    return ret;
}

void *CKeyHandler::make_keylist2(int *length)
{
    struct ssh2_userkey *key;
    int i, len, nkeys;
    unsigned char *blob, *p, *ret;
    int bloblen;

    /*
     * Count up the number and length of keys we hold.
     */
    len = 4;
    nkeys = 0;
    for (i = 0; NULL != (key = (ssh2_userkey*)index234(ssh2keys, i)); i++) {
		nkeys++;
		len += 4;	       /* length field */
		blob = key->alg->public_blob(key->data, &bloblen);
		len += bloblen;
		sfree(blob);
		len += 4 + strlen(key->comment);
    }

    /* Allocate the buffer. */
    p = ret = snewn(len, unsigned char);
    if (length) *length = len;

    /*
     * Packet header is the obvious five bytes, plus four
     * bytes for the key count.
     */
    PUT_32BIT(p, nkeys);
    p += 4;
    for (i = 0; NULL != (key = (ssh2_userkey*)index234(ssh2keys, i)); i++) {
		blob = key->alg->public_blob(key->data, &bloblen);
		PUT_32BIT(p, bloblen);
		p += 4;
		memcpy(p, blob, bloblen);
		p += bloblen;
		sfree(blob);
		PUT_32BIT(p, strlen(key->comment));
		memcpy(p + 4, key->comment, strlen(key->comment));
		p += 4 + strlen(key->comment);
    }

    //assert(p - ret == len);
    return ret;
}

void CKeyHandler::DeleteKeys()
{
	struct RSAKey *rkey;
	struct ssh2_userkey *skey;

	CSingleLock lock(&m_CritSection, TRUE);

	while((rkey = (RSAKey*)index234(rsakeys, 0)) != NULL) {
		del234(rsakeys, rkey);
		freersakey(rkey);
		sfree(rkey);
	}

	while ((skey = (ssh2_userkey*)index234(ssh2keys, 0)) != NULL) {
		del234(ssh2keys, skey);
		skey->alg->freekey(skey->data);
		sfree(skey);
	}
}

BOOL CKeyHandler::AddKey(const char *data, int len, const char *pass, CString &err)
{
	// TODO: lock more specific
	CSingleLock lock(&m_CritSection, TRUE);

    char passphrase[PASSPHRASE_MAXLEN];
    struct RSAKey *rkey = NULL;
    struct ssh2_userkey *skey = NULL;
    int ret;
    int attempts;
    const char *error = NULL;
    int type;

	type = key_type_data(data, len);
    if(type != SSH_KEYTYPE_SSH1 && type != SSH_KEYTYPE_SSH2) {
		err.Format("Wrong key format");
		return FALSE;
    }

    /*
     * See if the key is already loaded (in the primary Pageant,
     * which may or may not be us).
     */
    {
		void *blob;
		unsigned char *keylist, *p;
		int i, nkeys, bloblen, keylistlen;

		if (type == SSH_KEYTYPE_SSH1) {
			if (!rsakey_pubblob_data(data, len, &blob, &bloblen, NULL, &error)) {
				err.Format("Couldn't load private key (%s)", error);
				return FALSE;
			}
			keylist = (unsigned char*)make_keylist1(&keylistlen);
		} else {
			unsigned char *blob2;
			blob = ssh2_userkey_loadpub_data(data, len, NULL, &bloblen, NULL, &error);
			if (!blob) {
				err.Format("Couldn't load private key (%s)", error);
				return FALSE;
			}
			/* For our purposes we want the blob prefixed with its length */
			blob2 = snewn(bloblen+4, unsigned char);
			PUT_32BIT(blob2, bloblen);
			memcpy(blob2 + 4, blob, bloblen);
			sfree(blob);
			blob = blob2;

			keylist = (unsigned char*)make_keylist2(&keylistlen);
		}

		// blob is the key as char* (?)
		// keylist is the list of keys

		if(keylist) {
			if(keylistlen < 4) {
				err.Format("Received broken key list?!");
				return FALSE;
			}
			nkeys = GET_32BIT(keylist);
			p = keylist + 4;
			keylistlen -= 4;

			for (i = 0; i < nkeys; i++) {
				if (!memcmp(blob, p, bloblen)) {
					/* Key is already present; we can now leave. */
					sfree(keylist);
					sfree(blob);
					return TRUE;
				}
				/* Now skip over public blob */
				if (type == SSH_KEYTYPE_SSH1) {
					int n = rsa_public_blob_len(p, keylistlen);
					if (n < 0) {
						err.Format("Received broken key list?!");
						return FALSE;
					}
					p += n;
					keylistlen -= n;
				} else {
					int n;
					if (keylistlen < 4) {
						err.Format("Received broken key list?!");
						return FALSE;
					}
					n = 4 + GET_32BIT(p);
					if (keylistlen < n) {
						err.Format("Received broken key list?!");
						return FALSE;
					}
					p += n;
					keylistlen -= n;
				}
		
				/* Now skip over comment field */
				{
					int n;
					if (keylistlen < 4) {
						err.Format("Received broken key list?!");
						return FALSE;
					}
					n = 4 + GET_32BIT(p);
					if (keylistlen < n) {
						err.Format("Received broken key list?!");
						return FALSE;
					}
					p += n;
					keylistlen -= n;
				}
			}
			sfree(keylist);
		} // if(keylist)
		sfree(blob);
	} // See if the key is already loaded

	
    if(type == SSH_KEYTYPE_SSH1)
		rkey = snew(struct RSAKey);

	for(attempts = 0; attempts < 2; attempts++) {
		switch(attempts)
		{
			case 0:
				*passphrase = '\0';
				break;
			case 1:
				strcpy(passphrase, pass);
				break;
		}
		if(type == SSH_KEYTYPE_SSH1)
			ret = loadrsakey_data(data, len, rkey, passphrase, &error);
		else {
			skey = ssh2_load_userkey_data(data, len, passphrase, &error);
			if(skey == SSH2_WRONG_PASSPHRASE)
				ret = -1;
			else if (!skey)
				ret = 0;
			else
				ret = 1;
		}
		memset(passphrase, 0, PASSPHRASE_MAXLEN);
		if(ret != -1)
			break;
	}

	if(ret == -1)
		ret = 0;

    if (ret == 0) {
		err.Format("Could not load private key");
		if(type == SSH_KEYTYPE_SSH1)
			sfree(rkey);
		return FALSE;
    }

    if (type == SSH_KEYTYPE_SSH1) {
		if(add234(rsakeys, rkey) != rkey)
			sfree(rkey);	       /* already present, don't waste RAM */
    } else {
		if(add234(ssh2keys, skey) != skey) {
			skey->alg->freekey(skey->data);
			sfree(skey);	       /* already present, don't waste RAM */
		}
    }

	return TRUE;
}

void CKeyHandler::ListKeys(CKeyList &keys)
{
    struct RSAKey *rkey;
    struct ssh2_userkey *skey;
    int i;

		for (i = 0; NULL != (rkey = (RSAKey*)index234(rsakeys, i)); i++) {
			char listentry[512], *p;
			/*
			* Replace two spaces in the fingerprint with tabs, for
			* nice alignment in the box.
			*/
			strcpy(listentry, "ssh1\t");
			p = listentry + strlen(listentry);
			rsa_fingerprint(p, sizeof(listentry) - (p - listentry), rkey);
			p = strchr(listentry, ' ');
			if (p)
			*p = '\t';
			p = strchr(listentry, ' ');
			if (p)
			*p = '\t';

			keys.Add(CString(listentry));
		}
		for (i = 0; NULL != (skey = (ssh2_userkey*)index234(ssh2keys, i)); i++) {
			char listentry[512], *p;
			int len;
			/*
			* Replace two spaces in the fingerprint with tabs, for
			* nice alignment in the box.
			*/
			p = skey->alg->fingerprint(skey->data);
			strncpy(listentry, p, sizeof(listentry));
			p = strchr(listentry, ' ');
			if (p)
			*p = '\t';
			p = strchr(listentry, ' ');
			if (p)
			*p = '\t';
			len = strlen(listentry);
			if (len < sizeof(listentry) - 2) {
			listentry[len] = '\t';
			strncpy(listentry + len + 1, skey->comment,
				sizeof(listentry) - len - 1);
			}
			keys.Add(CString(listentry));
		}
}

int CKeyHandler::MessageHandler(unsigned char *msg)
{
	// TODO: make locking more specific
	CSingleLock lock(&m_CritSection, TRUE);

	int retval = -1;

    unsigned char *p = msg;
    unsigned char *ret = msg;
    unsigned char *msgend;
    int type;

    /*
     * Get the message length.
     */
    msgend = p + 4 + GET_32BIT(p);

    /*
     * Get the message type.
     */
    if (msgend < p+5)
		goto failure;
    type = p[4];

    p += 5;
    switch (type) {
      case SSH1_AGENTC_REQUEST_RSA_IDENTITIES:
	/*
	 * Reply with SSH1_AGENT_RSA_IDENTITIES_ANSWER.
	 */
	{
	    int len;
	    void *keylist;

	    ret[4] = SSH1_AGENT_RSA_IDENTITIES_ANSWER;
	    keylist = make_keylist1(&len);
	    if (len + 5 > AGENT_MAX_MSGLEN) {
			sfree(keylist);
			goto failure;
	    }
	    PUT_32BIT(ret, len + 1);
	    memcpy(ret + 5, keylist, len);
	    sfree(keylist);

		retval = len + 5;
	}
	break;
      case SSH2_AGENTC_REQUEST_IDENTITIES:
	/*
	 * Reply with SSH2_AGENT_IDENTITIES_ANSWER.
	 */
	{
	    int len;
	    void *keylist;

	    ret[4] = SSH2_AGENT_IDENTITIES_ANSWER;
	    keylist = make_keylist2(&len);
	    if (len + 5 > AGENT_MAX_MSGLEN) {
			sfree(keylist);
			goto failure;
	    }
	    PUT_32BIT(ret, len + 1);
	    memcpy(ret + 5, keylist, len);
	    sfree(keylist);
		retval = len + 5;
	}
	break;
      case SSH1_AGENTC_RSA_CHALLENGE:
	/*
	 * Reply with either SSH1_AGENT_RSA_RESPONSE or
	 * SSH_AGENT_FAILURE, depending on whether we have that key
	 * or not.
	 */
	{
	    struct RSAKey reqkey, *key;
	    Bignum challenge, response;
	    unsigned char response_source[48], response_md5[16];
	    struct MD5Context md5c;
	    int i, len;

	    p += 4;
	    i = ssh1_read_bignum(p, msgend - p, &reqkey.exponent);
	    if (i < 0)
		goto failure;
	    p += i;
	    i = ssh1_read_bignum(p, msgend - p, &reqkey.modulus);
	    if (i < 0)
		goto failure;
	    p += i;
	    i = ssh1_read_bignum(p, msgend - p, &challenge);
	    if (i < 0)
		goto failure;
	    p += i;
	    if (msgend < p+16) {
		freebn(reqkey.exponent);
		freebn(reqkey.modulus);
		freebn(challenge);
		goto failure;
	    }
	    memcpy(response_source + 32, p, 16);
	    p += 16;
	    if (msgend < p+4 ||
		GET_32BIT(p) != 1 ||
		(key = (RSAKey*)find234(rsakeys, &reqkey, NULL)) == NULL) {
		freebn(reqkey.exponent);
		freebn(reqkey.modulus);
		freebn(challenge);
		goto failure;
	    }
	    response = rsadecrypt(challenge, key);
	    for (i = 0; i < 32; i++)
		response_source[i] = bignum_byte(response, 31 - i);

	    MD5Init(&md5c);
	    MD5Update(&md5c, response_source, 48);
	    MD5Final(response_md5, &md5c);
	    memset(response_source, 0, 48);	/* burn the evidence */
	    freebn(response);	       /* and that evidence */
	    freebn(challenge);	       /* yes, and that evidence */
	    freebn(reqkey.exponent);   /* and free some memory ... */
	    freebn(reqkey.modulus);    /* ... while we're at it. */

	    /*
	     * Packet is the obvious five byte header, plus sixteen
	     * bytes of MD5.
	     */
	    len = 5 + 16;
	    PUT_32BIT(ret, len - 4);
	    ret[4] = SSH1_AGENT_RSA_RESPONSE;
	    memcpy(ret + 5, response_md5, 16);
		retval = len;
	}
	break;
      case SSH2_AGENTC_SIGN_REQUEST:
	/*
	 * Reply with either SSH2_AGENT_SIGN_RESPONSE or
	 * SSH_AGENT_FAILURE, depending on whether we have that key
	 * or not.
	 */
	{
	    struct ssh2_userkey *key;
	    struct BLOBSTRUCT b;
	    unsigned char *data, *signature;
	    int datalen, siglen, len;

	    if (msgend < p+4)
			goto failure;
	    b.len = GET_32BIT(p);
	    p += 4;
	    if (msgend < p+b.len)
			goto failure;
	    b.blob = p;
	    p += b.len;
	    if (msgend < p+4)
			goto failure;
	    datalen = GET_32BIT(p);
	    p += 4;
	    if (msgend < p+datalen)
			goto failure;
	    data = p;
	    key = (ssh2_userkey*)find234(ssh2keys, &b, cmpkeys_ssh2_asymm);
	    if (!key)
			goto failure;
	    signature = key->alg->sign(key->data, (char*)data, datalen, &siglen);
	    len = 5 + 4 + siglen;
	    PUT_32BIT(ret, len - 4);
	    ret[4] = SSH2_AGENT_SIGN_RESPONSE;
	    PUT_32BIT(ret + 5, siglen);
	    memcpy(ret + 5 + 4, signature, siglen);
	    sfree(signature);
		retval = len;
	}
	break;
      case SSH1_AGENTC_ADD_RSA_IDENTITY:
	/*
	 * Add to the list and return SSH_AGENT_SUCCESS, or
	 * SSH_AGENT_FAILURE if the key was malformed.
	 */
	{
	    struct RSAKey *key;
	    char *comment;
            int n, commentlen;

	    key = snew(struct RSAKey);
	    memset(key, 0, sizeof(struct RSAKey));

	    n = makekey(p, msgend - p, key, NULL, 1);
	    if (n < 0) {
		freersakey(key);
		sfree(key);
		goto failure;
	    }
	    p += n;

	    n = makeprivate(p, msgend - p, key);
	    if (n < 0) {
		freersakey(key);
		sfree(key);
		goto failure;
	    }
	    p += n;

	    n = ssh1_read_bignum(p, msgend - p, &key->iqmp);  /* p^-1 mod q */
	    if (n < 0) {
		freersakey(key);
		sfree(key);
		goto failure;
	    }
	    p += n;

	    n = ssh1_read_bignum(p, msgend - p, &key->p);  /* p */
	    if (n < 0) {
		freersakey(key);
		sfree(key);
		goto failure;
	    }
	    p += n;

	    n = ssh1_read_bignum(p, msgend - p, &key->q);  /* q */
	    if (n < 0) {
		freersakey(key);
		sfree(key);
		goto failure;
	    }
	    p += n;

	    if (msgend < p+4) {
		freersakey(key);
		sfree(key);
		goto failure;
	    }
            commentlen = GET_32BIT(p);

	    if (msgend < p+commentlen) {
		freersakey(key);
		sfree(key);
		goto failure;
	    }

	    comment = snewn(commentlen+1, char);
	    if (comment) {
		memcpy(comment, p + 4, commentlen);
                comment[commentlen] = '\0';
		key->comment = comment;
	    }
	    PUT_32BIT(ret, 1);
	    ret[4] = SSH_AGENT_FAILURE;
	    if (add234(rsakeys, key) == key) {
			ret[4] = SSH_AGENT_SUCCESS;
	    } else {
		freersakey(key);
		sfree(key);
	    }
		retval = 5;
	}
	break;
      case SSH2_AGENTC_ADD_IDENTITY:
	/*
	 * Add to the list and return SSH_AGENT_SUCCESS, or
	 * SSH_AGENT_FAILURE if the key was malformed.
	 */
	{
	    struct ssh2_userkey *key;
	    char *comment, *alg;
	    int alglen, commlen;
	    int bloblen;


	    if(msgend < p+4)
			goto failure;
	    alglen = GET_32BIT(p);
		p += 4;
	    if(msgend < p+alglen)
			goto failure;
	    alg = (char*)p;
	    p += alglen;

	    key = snew(struct ssh2_userkey);
	    /* Add further algorithm names here. */
	    if (alglen == 7 && !memcmp(alg, "ssh-rsa", 7))
			key->alg = &ssh_rsa;
	    else if (alglen == 7 && !memcmp(alg, "ssh-dss", 7))
			key->alg = &ssh_dss;
	    else {
			sfree(key);
			goto failure;
	    }

	    bloblen = msgend - p;
	    key->data = key->alg->openssh_createkey(&p, &bloblen);
	    if (!key->data) {
			sfree(key);
			goto failure;
	    }

	    /*
	     * p has been advanced by openssh_createkey, but
	     * certainly not _beyond_ the end of the buffer.
	     */
	    //assert(p <= msgend);

	    if (msgend < p+4) {
			key->alg->freekey(key->data);
			sfree(key);
			goto failure;
	    }
	    commlen = GET_32BIT(p);
	    p += 4;

	    if (msgend < p+commlen) {
			key->alg->freekey(key->data);
			sfree(key);
			goto failure;
	    }
	    comment = snewn(commlen + 1, char);
	    if (comment) {
			memcpy(comment, p, commlen);
			comment[commlen] = '\0';
	    }
	    key->comment = comment;

	    PUT_32BIT(ret, 1);
	    ret[4] = SSH_AGENT_FAILURE;
	    if (add234(ssh2keys, key) == key) {
			ret[4] = SSH_AGENT_SUCCESS;
	    } else {
			key->alg->freekey(key->data);
			sfree(key->comment);
			sfree(key);
	    }
		retval = 5;
	}
	break;
      case SSH1_AGENTC_REMOVE_RSA_IDENTITY:
	/*
	 * Remove from the list and return SSH_AGENT_SUCCESS, or
	 * perhaps SSH_AGENT_FAILURE if it wasn't in the list to
	 * start with.
	 */
	{
	    struct RSAKey reqkey, *key;
	    int n;

	    n = makekey(p, msgend - p, &reqkey, NULL, 0);
	    if (n < 0)
		goto failure;

	    key = (RSAKey*)find234(rsakeys, &reqkey, NULL);
	    freebn(reqkey.exponent);
	    freebn(reqkey.modulus);
	    PUT_32BIT(ret, 1);
	    ret[4] = SSH_AGENT_FAILURE;
	    if (key) {
		del234(rsakeys, key);
		freersakey(key);
		sfree(key);
		ret[4] = SSH_AGENT_SUCCESS;
	    }
		retval = 5;
	}
	break;
      case SSH2_AGENTC_REMOVE_IDENTITY:
	/*
	 * Remove from the list and return SSH_AGENT_SUCCESS, or
	 * perhaps SSH_AGENT_FAILURE if it wasn't in the list to
	 * start with.
	 */
	{
	    struct ssh2_userkey *key;
	    struct BLOBSTRUCT b;

	    if (msgend < p+4)
		goto failure;
	    b.len = GET_32BIT(p);
	    p += 4;

	    if (msgend < p+b.len)
		goto failure;
	    b.blob = p;
	    p += b.len;

	    key = (ssh2_userkey*)find234(ssh2keys, &b, cmpkeys_ssh2_asymm);
	    if (!key)
		goto failure;

	    PUT_32BIT(ret, 1);
	    ret[4] = SSH_AGENT_FAILURE;
	    if (key) {
		del234(ssh2keys, key);
		key->alg->freekey(key->data);
		sfree(key);
		ret[4] = SSH_AGENT_SUCCESS;
	    }
		retval = 5;
	}
	break;
      case SSH1_AGENTC_REMOVE_ALL_RSA_IDENTITIES:
	/*
	 * Remove all SSH-1 keys. Always returns success.
	 */
	{
	    struct RSAKey *rkey;

	    while ((rkey = (RSAKey*)index234(rsakeys, 0)) != NULL) {
		del234(rsakeys, rkey);
		freersakey(rkey);
		sfree(rkey);
	    }

	    PUT_32BIT(ret, 1);
	    ret[4] = SSH_AGENT_SUCCESS;
		retval = 5;
	}
	break;
      case SSH2_AGENTC_REMOVE_ALL_IDENTITIES:
	/*
	 * Remove all SSH-2 keys. Always returns success.
	 */
	{
	    struct ssh2_userkey *skey;

	    while ((skey = (ssh2_userkey*)index234(ssh2keys, 0)) != NULL) {
		del234(ssh2keys, skey);
		skey->alg->freekey(skey->data);
		sfree(skey);
	    }

	    PUT_32BIT(ret, 1);
	    ret[4] = SSH_AGENT_SUCCESS;
		retval = 5;
	}
	break;
      default:
      failure:
	/*
	 * Unrecognised message. Return SSH_AGENT_FAILURE.
	 */
	PUT_32BIT(ret, 1);
	ret[4] = SSH_AGENT_FAILURE;
	retval = 5;
	break;
    }

	return retval;
}


