/*
  PuttyAgent plugin for KeePass 1
  Copyright (C) 2011 Nikolaus Hammler <nikolaus@hammler.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "StdAfx.h"
#include <assert.h>
#include "PuttyAgent.h"
#include "InfoDlg.h"
#include "ViewKeysDlg.h"

IKpAPI* g_pAPI = NULL;
HWND g_hKeePass = NULL;

KP_IMPL_STDREFIMPL(PuttyAgent)

KP_EXPORT HRESULT KP_API KP_I_INITIALIZELIB_DECL(IKpUnknown* pAPI)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	UNREFERENCED_PARAMETER(pAPI);
	return S_OK;
}

KP_EXPORT HRESULT KP_API KP_I_RELEASELIB_DECL(IKpUnknown* pAPI)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	UNREFERENCED_PARAMETER(pAPI);
	SAFE_RELEASE(g_pAPI);
	return S_OK;
}

KP_EXPORT HRESULT KP_API KP_I_CREATEINSTANCE_DECL(REFIID riid, void** ppvObject, IKpUnknown* pAPI)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	*ppvObject = NULL;

	if(pAPI == NULL) return E_UNEXPECTED; // Require API

	if((riid == IID_IKpUnknown) || (riid == IID_IKpPlugin))
	{
		if(pAPI->QueryInterface(IID_IKpAPI, (void**)&g_pAPI) != S_OK)
			return E_FAIL;

		IKpPlugin* pImpl = new PuttyAgent();
		*ppvObject = ((riid == IID_IKpPlugin) ? pImpl : (IKpUnknown*)pImpl);
		return S_OK;
	}

	return E_NOINTERFACE;
}

PuttyAgent::PuttyAgent()
	: m_pThread(0)
{
	KP_IMPL_CONSTRUCT;

	ZeroMemory(&m_menuItems[0], sizeof(KP_MENU_ITEM) * MENUITEMCOUNT);

	m_menuItems[0].lpCommandString = _T("PuttyAgent");
	m_menuItems[0].dwFlags = KPMIF_POPUP_START;

	m_menuItems[1].lpCommandString = _T("&Show Keys");
	m_menuItems[1].dwIcon = 34;
	
	m_menuItems[2].lpCommandString = _T("Enable PuttyAgent");
	m_menuItems[2].dwFlags = KPMIF_CHECKBOX;
	m_menuItems[2].dwState = 0;

	ASSERT( g_pAPI );
	g_hKeePass = g_pAPI->GetMainWindowHandle();

	StartAgent();
}

PuttyAgent::~PuttyAgent()
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	StopAgent();
}

void PuttyAgent::StartAgent()
{
	CWnd *agent = CWnd::FindWindow(_T("Pageant"), _T("Pageant"));
	if(agent) {
		MessageBox(NULL, _T("Pageant already running. Not enabling agent functionality"), _T("KeePass PuttyAgent plugin"), MB_OK | MB_ICONWARNING);
	}
	else {
		ASSERT( m_pThread == 0 );
		m_pThread = new CPageantThread();
		m_pThread->CreateThread();

		if(::WaitForSingleObject(m_pThread->m_evStartSuccessful.m_hObject, 5000) != WAIT_OBJECT_0)
		{
			MessageBox(NULL, _T("Pageant Thread has been started but thread did not respond for 5 seconds."), _T("Fatal Error"), MB_OK | MB_ICONERROR);
			StopAgent();
			return;
		}

		m_menuItems[2].dwState = KPMIS_CHECKED;
	}
}

BOOL PuttyAgent::StopAgent()
{
	if(m_pThread)
	{
		if(!m_pThread->StopThread())
		{
			m_pThread->KillThread();
		}
		delete m_pThread;
		m_pThread = 0;
		m_menuItems[2].dwState = 0;
		return TRUE;
	}
	return FALSE;
}

void PuttyAgent::UpdateKeys()
{
	IKpDatabase* p = NULL;

	if(!g_pAPI->IsFileOpen())
		return;

	if(!m_pThread) {
		return;
	}
	if(!m_pThread->IsRunning()) {
		return;
	}

	CKeyHandler::GetInstance().DeleteKeys();

	if(g_pAPI->QueryInstance(SCLSID_KpDatabase, IID_IKpDatabase, (void**)&p) == S_OK)
	{
		int count = p->GetEntryCount();
		for(int i=0; i < count; i++) {
			PW_ENTRY *entry = p->GetEntry(i);
			if(entry->uBinaryDataLen) {

				CString str = entry->pszBinaryDesc;

				if(str.Right(4) == _T(".ppk")) {
					p->UnlockEntryPassword(entry);
					CString password = entry->pszPassword;
					p->LockEntryPassword(entry);

					CString err;

					if(!CKeyHandler::GetInstance().AddKey((const char*)entry->pBinaryData, entry->uBinaryDataLen, password, err)) {
						CString title;
						title.Format(_T("Error while loading key %s (entry %s)"), entry->pszBinaryDesc, entry->pszTitle);
						MessageBox(NULL, err, title, MB_OK | MB_ICONWARNING);
					}

					char *p = password.GetBuffer(512);
					memset(p, 0, 512);
					password.ReleaseBufferSetLength(512);
				}
			}
		}
		p->Release();
	}
}

STDMETHODIMP PuttyAgent::QueryInterface(REFIID riid, void** ppvObject)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	KP_SUPPORT_INTERFACE(IID_IKpUnknown, IKpUnknown);
	KP_SUPPORT_INTERFACE(IID_IKpPlugin, IKpPlugin);

	*ppvObject = NULL;
	return E_NOINTERFACE;
}

STDMETHODIMP_(BOOL) PuttyAgent::OnMessage(DWORD dwCode, LPARAM lParamW, LPARAM lParamL)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	switch(dwCode)
	{
		case KPM_DIRECT_CONFIG:
			ShowSettings();
			break;
		case KPM_PLUGIN_INFO:
			{
				CInfoDlg dlg;
				dlg.DoModal();
			}
			break;
		case KPM_DIRECT_EXEC:
			if(lParamW == m_menuItems[2].dwCommandID)
			{
				if(m_menuItems[2].dwState == KPMIS_CHECKED)
				{
					StopAgent();
					break;
				}
				else
				{
					StartAgent();
					UpdateKeys();
					break;
				}
			}
			if(!m_pThread || !m_pThread->IsRunning()) {
				MessageBox(g_pAPI->GetMainWindowHandle(), _T("Agent functionality not enabled"), _T("PuttyAgent plugin"), MB_OK | MB_ICONASTERISK);
				break;
			}
			if(lParamW == m_menuItems[1].dwCommandID)
			{
				CKeyList keys;
				CKeyHandler::GetInstance().ListKeys(keys);

				CWnd *pMain = CWnd::FromHandle(g_pAPI->GetMainWindowHandle());
				CViewKeysDlg dlg(pMain, keys);
				dlg.DoModal();
			}
			break;
		case KPM_SAVEDB_POST:
		case KPM_OPENDB_POST:
			UpdateKeys();
			break;
	}
	return TRUE;
}

STDMETHODIMP_(LPCTSTR) PuttyAgent::GetProperty(LPCTSTR lpName)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	UNREFERENCED_PARAMETER(lpName);
	return NULL;
}

STDMETHODIMP PuttyAgent::SetProperty(LPCTSTR lpName, LPCTSTR lpValue)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	UNREFERENCED_PARAMETER(lpName);
	UNREFERENCED_PARAMETER(lpValue);
	return E_FAIL;
}

STDMETHODIMP_(DWORD) PuttyAgent::GetMenuItemCount()
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());
	return MENUITEMCOUNT;
}

STDMETHODIMP_(KP_MENU_ITEM*) PuttyAgent::GetMenuItems()
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());
	return m_menuItems;
}

BOOL CALLBACK SettingsDlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	IKpConfig *pConf = NULL;
	IKpUtilities *pUtils = NULL;
	TCHAR strSockPath[SI_REGSIZE] = _T("");

	switch(uMsg)
	{
		case WM_INITDIALOG:
			if(g_pAPI->CreateInstance(CLSID_KpConfig_ReadOnly,
				IID_IKpConfig, (void**)&pConf) != S_OK)
				return FALSE;

			assert( pConf );

			CheckDlgButton(hwndDlg, IDC_BALLOON_REQUEST,
				pConf->GetBool(CONF_BALOON_REQ, CONF_BALOON_REQ_DEF) ? BST_CHECKED : BST_UNCHECKED);

			CheckDlgButton(hwndDlg, IDC_BALLOON_INFO,
				pConf->GetBool(CONF_BALOON_INFO, CONF_BALOON_INFO_DEF) ? BST_CHECKED : BST_UNCHECKED);

			{
				BOOL bCygEnabled = pConf->GetBool(CONF_CYG_ENABLE, CONF_CYG_ENABLE_DEF);
				CheckDlgButton(hwndDlg, IDC_CYG_ENABLE, bCygEnabled ? BST_CHECKED : BST_UNCHECKED);

				if(pConf->Get(CONF_CYG_SOCKPATH, strSockPath))
					SetDlgItemText(hwndDlg, IDC_CYG_SOCKPATH, strSockPath);

				if(!bCygEnabled)
					EnableWindow(GetDlgItem(hwndDlg, IDC_CYG_SOCKPATH), FALSE);
			}

			pConf->Release();
			return TRUE;
		case WM_COMMAND:
			switch(wParam)
			{
				case IDCANCEL:
					EndDialog(hwndDlg, 0);
					return TRUE;
				case IDC_CYG_ENABLE:
					if(IsDlgButtonChecked(hwndDlg, IDC_CYG_ENABLE) == BST_CHECKED)
						EnableWindow(GetDlgItem(hwndDlg, IDC_CYG_SOCKPATH), TRUE);
					else
						EnableWindow(GetDlgItem(hwndDlg, IDC_CYG_SOCKPATH), FALSE);
					return TRUE;
				case IDC_SOCKBROWSE:
					if(g_pAPI->QueryInstance(SCLSID_KpUtilities, IID_IKpUtilities, (void **)&pUtils) != S_OK)
						return FALSE;
					assert( pUtils );

					GetDlgItemText(hwndDlg, IDC_CYG_SOCKPATH, strSockPath, SI_REGSIZE-1);
					pUtils->ShowFileDialog(FALSE, _T("*"), strSockPath, SI_REGSIZE-1);
					SetDlgItemText(hwndDlg, IDC_CYG_SOCKPATH, strSockPath);

					pUtils->Release();
					return TRUE;
				case IDOK:
					if(g_pAPI->CreateInstance(CLSID_KpConfig,
						IID_IKpConfig, (void**)&pConf) != S_OK)
						return FALSE;

					assert( pConf );

					pConf->SetBool(CONF_BALOON_REQ, IsDlgButtonChecked(hwndDlg, IDC_BALLOON_REQUEST) == BST_CHECKED ? TRUE : FALSE);
					pConf->SetBool(CONF_BALOON_INFO, IsDlgButtonChecked(hwndDlg, IDC_BALLOON_INFO) == BST_CHECKED ? TRUE : FALSE);

					pConf->SetBool(CONF_CYG_ENABLE, IsDlgButtonChecked(hwndDlg, IDC_CYG_ENABLE) == BST_CHECKED ? TRUE : FALSE);

					if(GetDlgItemText(hwndDlg, IDC_CYG_SOCKPATH, strSockPath, SI_REGSIZE-1) != 0)
						pConf->Set(CONF_CYG_SOCKPATH, strSockPath);

					pConf->Release();
					EndDialog(hwndDlg, 0);
					return TRUE;
			}
	}
	return FALSE;
}

void PuttyAgent::ShowSettings(void)
{
	assert( g_pAPI );
	DialogBox(GetModuleHandle(_T("PuttyAgent.dll")), MAKEINTRESOURCE(IDD_SETTINGS), g_pAPI->GetMainWindowHandle(), SettingsDlgProc);
	// restart agent only if was started before
	if(StopAgent())
		StartAgent();
}
