/*
  PuttyAgent plugin for KeePass 1
  Copyright (C) 2011 Nikolaus Hammler <nikolaus@hammler.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*
  The code in this file is derived from SSHPUBK.C of Pageants sources and
  modified such that it reads key files from memory rather than from files.

  The modified functions are postfixed by _data
*/

#include <stdlib.h>

#include "putty_additional.h"
#include "putty/ssh.h"

#define rsa_signature "SSH PRIVATE KEY FILE FORMAT 1.1\n"

#include <windows.h>

/*
 * We need this to link with the RSA code, because rsaencrypt()
 * pads its data with random bytes. Since we only use rsadecrypt()
 * and the signing functions, which are deterministic, this should
 * never be called.
 *
 * If it _is_ called, there is a _serious_ problem, because it
 * won't generate true random numbers. So we must scream, panic,
 * and exit immediately if that should happen.
 */
int random_byte(void)
{
    MessageBox(NULL, "Internal Error", "PuttyAgent plugin", MB_OK | MB_ICONERROR);
    exit(0);
    /* this line can't be reached but it placates MSVC's warnings :-) */
    return 0;
}

/*
 * Warn about the obsolescent key file format.
 * Wird von SSH libs benoetigt
 */
void old_keyfile_warning(void)
{
    static const char mbtitle[] = "PuTTY Key File Warning";
    static const char message[] =
	"You are loading an SSH-2 private key which has an\n"
	"old version of the file format. This means your key\n"
	"file is not fully tamperproof. Future versions of\n"
	"PuTTY may stop supporting this private key format,\n"
	"so we recommend you convert your key to the new\n"
	"format.\n"
	"\n"
	"You can perform this conversion by loading the key\n"
	"into PuTTYgen and then saving it again.";

    MessageBox(NULL, message, mbtitle, MB_OK);
}

/*
 * Print a modal (Really Bad) message box and perform a fatal exit.
 */
void modalfatalbox(char *fmt, ...)
{
    va_list ap;
    char *buf;

    va_start(ap, fmt);
    buf = dupvprintf(fmt, ap);
    va_end(ap);
    MessageBox(NULL, buf, "Pageant Fatal Error", MB_SYSTEMMODAL | MB_ICONERROR | MB_OK);
    sfree(buf);
    exit(1);
}

int key_type_data(const char *buf, int len)
{
    const char putty2_sig[] = "PuTTY-User-Key-File-";
    const char sshcom_sig[] = "---- BEGIN SSH2 ENCRYPTED PRIVAT";
    const char openssh_sig[] = "-----BEGIN ";

	if(len < 32)
		return SSH_KEYTYPE_UNOPENABLE;

    if (!memcmp(buf, rsa_signature, sizeof(rsa_signature)-1))
		return SSH_KEYTYPE_SSH1;
    if (!memcmp(buf, putty2_sig, sizeof(putty2_sig)-1))
		return SSH_KEYTYPE_SSH2;
    if (!memcmp(buf, openssh_sig, sizeof(openssh_sig)-1))
		return SSH_KEYTYPE_OPENSSH;
    if (!memcmp(buf, sshcom_sig, sizeof(sshcom_sig)-1))
		return SSH_KEYTYPE_SSHCOM;
    return SSH_KEYTYPE_UNKNOWN;	       /* unrecognised or EOF */
}

static int loadrsakey_main_data(const char *buf, int len, struct RSAKey *key, int pub_only,
			   char **commentptr, char *passphrase,
			   const char **error)
{
    //unsigned char buf[16384];
    unsigned char keybuf[16];
    //int len;
    int i, j, ciphertype;
    int ret = 0;
    struct MD5Context md5c;
    char *comment;

    *error = NULL;

    /* Slurp the whole file (minus the header) into a buffer. */
    //len = fread(buf, 1, sizeof(buf), fp);
    //fclose(fp);
    if (len < 0 || len == sizeof(buf)) {
		*error = "error reading file";
		goto end;		       /* file too big or not read */
    }

    i = 0;
    *error = "file format error";

    /*
     * A zero byte. (The signature includes a terminating NUL.)
     */
    if (len - i < 1 || buf[i] != 0)
	goto end;
    i++;

    /* One byte giving encryption type, and one reserved uint32. */
    if (len - i < 1)
	goto end;
    ciphertype = buf[i];
    if (ciphertype != 0 && ciphertype != SSH_CIPHER_3DES)
	goto end;
    i++;
    if (len - i < 4)
	goto end;		       /* reserved field not present */
    if (buf[i] != 0 || buf[i + 1] != 0 || buf[i + 2] != 0
	|| buf[i + 3] != 0) goto end;  /* reserved field nonzero, panic! */
    i += 4;

    /* Now the serious stuff. An ordinary SSH-1 public key. */
    i += makekey(buf + i, len, key, NULL, 1);
    if (i < 0)
	goto end;		       /* overran */

    /* Next, the comment field. */
    j = GET_32BIT(buf + i);
    i += 4;
    if (len - i < j)
	goto end;
    comment = snewn(j + 1, char);
    if (comment) {
	memcpy(comment, buf + i, j);
	comment[j] = '\0';
    }
    i += j;
    if (commentptr)
	*commentptr = dupstr(comment);
    if (key)
	key->comment = comment;
    else
	sfree(comment);

    if (pub_only) {
	ret = 1;
	goto end;
    }

    if (!key) {
	ret = ciphertype != 0;
	*error = NULL;
	goto end;
    }

    /*
     * Decrypt remainder of buffer.
     */
    if (ciphertype) {
	MD5Init(&md5c);
	MD5Update(&md5c, (unsigned char *)passphrase, strlen(passphrase));
	MD5Final(keybuf, &md5c);
	des3_decrypt_pubkey(keybuf, buf + i, (len - i + 7) & ~7);
	memset(keybuf, 0, sizeof(keybuf));	/* burn the evidence */
    }

    /*
     * We are now in the secret part of the key. The first four
     * bytes should be of the form a, b, a, b.
     */
    if (len - i < 4)
	goto end;
    if (buf[i] != buf[i + 2] || buf[i + 1] != buf[i + 3]) {
	*error = "wrong passphrase";
	ret = -1;
	goto end;
    }
    i += 4;

    /*
     * After that, we have one further bignum which is our
     * decryption exponent, and then the three auxiliary values
     * (iqmp, q, p).
     */
    j = makeprivate(buf + i, len - i, key);
    if (j < 0) goto end;
    i += j;
    j = ssh1_read_bignum(buf + i, len - i, &key->iqmp);
    if (j < 0) goto end;
    i += j;
    j = ssh1_read_bignum(buf + i, len - i, &key->q);
    if (j < 0) goto end;
    i += j;
    j = ssh1_read_bignum(buf + i, len - i, &key->p);
    if (j < 0) goto end;
    i += j;

    if (!rsa_verify(key)) {
	*error = "rsa_verify failed";
	freersakey(key);
	ret = 0;
    } else
	ret = 1;

  end:
    memset(buf, 0, sizeof(buf));       /* burn the evidence */
    return ret;

}

int loadrsakey_data(const char *buf, int len, struct RSAKey *key, char *passphrase, const char **errorstr)
{
    //FILE *fp;
    //char buf[64];
    int ret = 0;
    const char *error = NULL;

    //fp = f_open(*filename, "rb", FALSE);
    //if (!fp) {
	//error = "can't open file";
	//goto end;
    //}

	if(len < strlen(rsa_signature))
		return 0;

    /*
     * Read the first line of the file and see if it's a v1 private
     * key file.
     */
    //if (fgets(buf, sizeof(buf), fp) && !strcmp(buf, rsa_signature)) {
	if(!strncmp(buf, rsa_signature, strlen(rsa_signature))) {
		ret = loadrsakey_main_data(buf, len, key, FALSE, NULL, passphrase, &error);
		//fp = NULL;
		goto end;
    }

    /*
     * Otherwise, we have nothing. Return empty-handed.
     */
    error = "not an SSH-1 RSA file";

  end:
    //if (fp)
	//fclose(fp);
    if ((ret != 1) && errorstr)
	*errorstr = error;
    return ret;
}

/*
 * Return a malloc'ed chunk of memory containing the public blob of
 * an RSA key, as given in the agent protocol (modulus bits,
 * exponent, modulus).
 */
int rsakey_pubblob_data(const char *data, int len, void **blob, int *bloblen, char **commentptr, const char **errorstr)
{
    struct RSAKey key;
    int ret;
    const char *error = NULL;

    /* Default return if we fail. */
    *blob = NULL;
    *bloblen = 0;
    ret = 0;

	if(len < strlen(rsa_signature))
		return 0;

    /*
     * Read the first line of the file and see if it's a v1 private
     * key file.
     */
    if(!strncmp(data, rsa_signature, strlen(rsa_signature)))
	{
		memset(&key, 0, sizeof(key));
		if (loadrsakey_main_data(data, len, &key, TRUE, commentptr, NULL, &error)) {
			*blob = rsa_public_blob(&key, bloblen);
			freersakey(&key);
			ret = 1;
		}
	} else {
		error = "not an SSH-1 RSA file";
	}

    if ((ret != 1) && errorstr)
		*errorstr = error;
    return ret;
}

static int read_header_data(const char *data, int olen, char *header, int *curpos)
{
    int len = 39;
    int c;

	if(olen < 39)
		return 0;

    while (len > 0) {
		// c = fgetc(fp);
		c = data[(*curpos)++];

		//if (c == '\n' || c == '\r' || c == EOF)
		if (c == '\n' || c == '\r' || (*curpos) >= olen)
			return 0;		       /* failure */
		if (c == ':') {
			//c = fgetc(fp);
			c = data[(*curpos)++];
			if (c != ' ')
				return 0;
			*header = '\0';
			return 1;		       /* success! */
		}
		if (len == 0)
			return 0;		       /* failure */
		*header++ = c;
		len--;
    }
    return 0;			       /* failure */
}

static char *read_body_data(const char *data, int olen, int *curpos)
{
    char *text;
    int len;
    int size;
    int c;

    size = 128;
    text = snewn(size, char);
    len = 0;
    text[len] = '\0';

    while (1) {
		//c = fgetc(fp);
		c = data[(*curpos)++];
		if (c == '\r' || c == '\n') {
			//c = fgetc(fp);
			c = data[(*curpos)++];
			//if (c != '\r' && c != '\n' && c != EOF)
			//	ungetc(c, fp);
			if (c != '\r' && c != '\n' && (*curpos) <= olen)
				(*curpos)--;
			return text;
		}
		//if (c == EOF) {
		if((*curpos) > olen) {
			sfree(text);
			return NULL;
		}
		if (len + 1 >= size) {
			size += 128;
			text = sresize(text, size, char);
		}
		text[len++] = c;
		text[len] = '\0';
    }
}

static unsigned char *read_blob_data(const char *data, int olen, int nlines, int *bloblen, int *curpos)
{
    unsigned char *blob;
    char *line;
    int linelen, len;
    int i, j, k;

    /* We expect at most 64 base64 characters, ie 48 real bytes, per line. */
    blob = snewn(48 * nlines, unsigned char);
    len = 0;
    for (i = 0; i < nlines; i++) {
		//line = read_body(fp);
		line = read_body_data(data, olen, curpos);
		if (!line) {
			sfree(blob);
			return NULL;
		}
		linelen = strlen(line);
		if (linelen % 4 != 0 || linelen > 64) {
			sfree(blob);
			sfree(line);
			return NULL;
		}
		for (j = 0; j < linelen; j += 4) {
			k = base64_decode_atom(line + j, blob + len);
			if (!k) {
				sfree(line);
				sfree(blob);
				return NULL;
			}
			len += k;
		}
		sfree(line);
    }
    *bloblen = len;
    return blob;
}

unsigned char *ssh2_userkey_loadpub_data(const char *data, int len, char **algorithm,
				    int *pub_blob_len, char **commentptr,
				    const char **errorstr)
{
    //FILE *fp;
	int curpos = 0;
    char header[40], *b;
    const struct ssh_signkey *alg;
    unsigned char *public_blob;
    int public_blob_len;
    int i;
    const char *error = NULL;
    char *comment;

    public_blob = NULL;

    //fp = f_open(*filename, "rb", FALSE);
    //if (!fp) {
	//error = "can't open file";
	//goto error;
    //}

    /* Read the first header line which contains the key type. */
    if (!read_header_data(data, len, header, &curpos)
		|| (0 != strcmp(header, "PuTTY-User-Key-File-2") &&
			0 != strcmp(header, "PuTTY-User-Key-File-1"))) {
		error = "not a PuTTY SSH-2 private key";
		goto error;
    }

    error = "file format error";
    if ((b = read_body_data(data, len, &curpos)) == NULL)
		goto error;

    /* Select key algorithm structure. */
    alg = find_pubkey_alg(b);
    if (!alg) {
		sfree(b);
		goto error;
    }
    sfree(b);

    /* Read the Encryption header line. */
    if (!read_header_data(data, len, header, &curpos) || 0 != strcmp(header, "Encryption"))
		goto error;

    if ((b = read_body_data(data, len, &curpos)) == NULL)
		goto error;
    sfree(b);			       /* we don't care */

    /* Read the Comment header line. */
    if (!read_header_data(data, len, header, &curpos) || 0 != strcmp(header, "Comment"))
		goto error;
    if ((comment = read_body_data(data, len, &curpos)) == NULL)
		goto error;

    if (commentptr)
		*commentptr = comment;
    else
		sfree(comment);

    /* Read the Public-Lines header line and the public blob. */
    if (!read_header_data(data, len, header, &curpos) || 0 != strcmp(header, "Public-Lines"))
		goto error;
    if ((b = read_body_data(data, len, &curpos)) == NULL)
		goto error;

    i = atoi(b);

    sfree(b);
    if ((public_blob = read_blob_data(data, len, i, &public_blob_len, &curpos)) == NULL)
		goto error;

   // fclose(fp);
    if (pub_blob_len)
		*pub_blob_len = public_blob_len;
    if (algorithm)
		*algorithm = alg->name;
    return public_blob;

    /*
     * Error processing.
     */
  error:
    //if (fp)
	//fclose(fp);
    if (public_blob)
	sfree(public_blob);
    if (errorstr)
	*errorstr = error;
    return NULL;
}

struct ssh2_userkey *ssh2_load_userkey_data(const char *data, int olen, char *passphrase, const char **errorstr)
{
    //FILE *fp;
    char header[40], *b, *encryption, *comment, *mac;
    const struct ssh_signkey *alg;
    struct ssh2_userkey *ret;
    int cipher, cipherblk;
    unsigned char *public_blob, *private_blob;
    int public_blob_len, private_blob_len;
    int i, is_mac, old_fmt;
    int passlen = passphrase ? strlen(passphrase) : 0;
    const char *error = NULL;
	int curpos = 0;

    ret = NULL;			       /* return NULL for most errors */
    encryption = comment = mac = NULL;
    public_blob = private_blob = NULL;

    //fp = f_open(*filename, "rb", FALSE);
    //if (!fp) {
	//error = "can't open file";
	//goto error;
    //}

    /* Read the first header line which contains the key type. */
    if (!read_header_data(data, olen, header, &curpos))
		goto error;
    if (0 == strcmp(header, "PuTTY-User-Key-File-2")) {
		old_fmt = 0;
    } else if (0 == strcmp(header, "PuTTY-User-Key-File-1")) {
		/* this is an old key file; warn and then continue */
		old_keyfile_warning();
		old_fmt = 1;
    } else {
		error = "not a PuTTY SSH-2 private key";
		goto error;
    }
    error = "file format error";
    if ((b = read_body_data(data, olen, &curpos)) == NULL)
		goto error;
    /* Select key algorithm structure. */
    alg = find_pubkey_alg(b);
    if (!alg) {
		sfree(b);
		goto error;
    }
    sfree(b);

    /* Read the Encryption header line. */
    if (!read_header_data(data, olen, header, &curpos) || 0 != strcmp(header, "Encryption"))
		goto error;
    if ((encryption = read_body_data(data, olen, &curpos)) == NULL)
		goto error;
    if (!strcmp(encryption, "aes256-cbc")) {
		cipher = 1;
		cipherblk = 16;
    } else if (!strcmp(encryption, "none")) {
		cipher = 0;
		cipherblk = 1;
    } else {
		sfree(encryption);
		goto error;
    }

    /* Read the Comment header line. */
    if (!read_header_data(data, olen, header, &curpos) || 0 != strcmp(header, "Comment"))
		goto error;
    if ((comment = read_body_data(data, olen, &curpos)) == NULL)
		goto error;

    /* Read the Public-Lines header line and the public blob. */
    if (!read_header_data(data, olen, header, &curpos) || 0 != strcmp(header, "Public-Lines"))
		goto error;
    if ((b = read_body_data(data, olen, &curpos)) == NULL)
		goto error;
    i = atoi(b);
    sfree(b);
    if ((public_blob = read_blob_data(data, olen, i, &public_blob_len, &curpos)) == NULL)
		goto error;

    /* Read the Private-Lines header line and the Private blob. */
    if (!read_header_data(data, olen, header, &curpos) || 0 != strcmp(header, "Private-Lines"))
		goto error;
    if ((b = read_body_data(data, olen, &curpos)) == NULL)
		goto error;
    i = atoi(b);
    sfree(b);
    if ((private_blob = read_blob_data(data, olen, i, &private_blob_len, &curpos)) == NULL)
		goto error;

    /* Read the Private-MAC or Private-Hash header line. */
    if (!read_header_data(data, olen, header, &curpos))
		goto error;
    if (0 == strcmp(header, "Private-MAC")) {
		if ((mac = read_body_data(data, olen, &curpos)) == NULL)
			goto error;
		is_mac = 1;
    } else if (0 == strcmp(header, "Private-Hash") && old_fmt) {
		if ((mac = read_body_data(data, olen, &curpos)) == NULL)
			goto error;
			is_mac = 0;
		} else
			goto error;

		//fclose(fp);
		//fp = NULL;

    /*
     * Decrypt the private blob.
     */
    if (cipher) {
	unsigned char key[40];
	SHA_State s;

	if (!passphrase)
	    goto error;
	if (private_blob_len % cipherblk)
	    goto error;

	SHA_Init(&s);
	SHA_Bytes(&s, "\0\0\0\0", 4);
	SHA_Bytes(&s, passphrase, passlen);
	SHA_Final(&s, key + 0);
	SHA_Init(&s);
	SHA_Bytes(&s, "\0\0\0\1", 4);
	SHA_Bytes(&s, passphrase, passlen);
	SHA_Final(&s, key + 20);
	aes256_decrypt_pubkey(key, private_blob, private_blob_len);
    }

    /*
     * Verify the MAC.
     */
    {
	char realmac[41];
	unsigned char binary[20];
	unsigned char *macdata;
	int maclen;
	int free_macdata;

	if (old_fmt) {
	    /* MAC (or hash) only covers the private blob. */
	    macdata = private_blob;
	    maclen = private_blob_len;
	    free_macdata = 0;
	} else {
	    unsigned char *p;
	    int namelen = strlen(alg->name);
	    int enclen = strlen(encryption);
	    int commlen = strlen(comment);
	    maclen = (4 + namelen +
		      4 + enclen +
		      4 + commlen +
		      4 + public_blob_len +
		      4 + private_blob_len);
	    macdata = snewn(maclen, unsigned char);
	    p = macdata;
#define DO_STR(s,len) PUT_32BIT(p,(len));memcpy(p+4,(s),(len));p+=4+(len)
	    DO_STR(alg->name, namelen);
	    DO_STR(encryption, enclen);
	    DO_STR(comment, commlen);
	    DO_STR(public_blob, public_blob_len);
	    DO_STR(private_blob, private_blob_len);

	    free_macdata = 1;
	}

	if (is_mac) {
	    SHA_State s;
	    unsigned char mackey[20];
	    char header[] = "putty-private-key-file-mac-key";

	    SHA_Init(&s);
	    SHA_Bytes(&s, header, sizeof(header)-1);
	    if (cipher && passphrase)
		SHA_Bytes(&s, passphrase, passlen);
	    SHA_Final(&s, mackey);

	    hmac_sha1_simple(mackey, 20, macdata, maclen, binary);

	    memset(mackey, 0, sizeof(mackey));
	    memset(&s, 0, sizeof(s));
	} else {
	    SHA_Simple(macdata, maclen, binary);
	}

	if (free_macdata) {
	    memset(macdata, 0, maclen);
	    sfree(macdata);
	}

	for (i = 0; i < 20; i++)
	    sprintf(realmac + 2 * i, "%02x", binary[i]);

	if (strcmp(mac, realmac)) {
	    /* An incorrect MAC is an unconditional Error if the key is
	     * unencrypted. Otherwise, it means Wrong Passphrase. */
	    if (cipher) {
		error = "wrong passphrase";
		ret = SSH2_WRONG_PASSPHRASE;
	    } else {
		error = "MAC failed";
		ret = NULL;
	    }
	    goto error;
	}
    }
    sfree(mac);

    /*
     * Create and return the key.
     */
    ret = snew(struct ssh2_userkey);
    ret->alg = alg;
    ret->comment = comment;
    ret->data = alg->createkey(public_blob, public_blob_len,
			       private_blob, private_blob_len);
    if (!ret->data) {
	sfree(ret->comment);
	sfree(ret);
	ret = NULL;
	error = "createkey failed";
	goto error;
    }
    sfree(public_blob);
    sfree(private_blob);
    sfree(encryption);
    if (errorstr)
	*errorstr = NULL;
    return ret;

    /*
     * Error processing.
     */
  error:
    //if (fp)
	//fclose(fp);
    if (comment)
	sfree(comment);
    if (encryption)
	sfree(encryption);
    if (mac)
	sfree(mac);
    if (public_blob)
	sfree(public_blob);
    if (private_blob)
	sfree(private_blob);
    if (errorstr)
	*errorstr = error;
    return ret;
}




